use proc_macro::TokenStream;
use quote::{format_ident, quote};
use syn::{parse_macro_input, DeriveInput};

use convert_case::{Case, Casing};

fn parse_fk(attrs: &[syn::Attribute]) -> bool {
    for attr in attrs {
        if attr.path.segments.len() == 1
            && attr.path.segments.last().unwrap().ident == "microrm_foreign"
        {
            return true;
        }
    }

    false
}

pub(crate) fn derive(tokens: TokenStream) -> TokenStream {
    let input = parse_macro_input!(tokens as DeriveInput);

    let microrm_ref = crate::parse_microrm_ref(&input.attrs);

    let struct_name = &input.ident;
    let enum_name = format_ident!("{}Columns", &input.ident);
    let id_name = format_ident!("{}ID", &input.ident);

    let table_name = format!("{}", struct_name).to_case(Case::Snake);

    let st = match input.data {
        syn::Data::Struct(st) => st,
        _ => panic!("Can only use derive(Entity) on structs!"),
    };
    let fields = match st.fields {
        syn::Fields::Named(fields) => fields,
        _ => panic!("Can only use derive(Entity) on non-unit structs with named fields!"),
    };

    let mut variants = Vec::new();
    let mut field_names = Vec::new();
    let mut field_numbers = Vec::new();
    let mut field_types = Vec::new();
    let mut value_references = Vec::new();

    let mut foreign_keys = Vec::new();
    let mut foreign_key_impls = Vec::new();

    let mut build_clauses = Vec::new();

    let mut index: usize = 0;

    for name in fields.named.iter() {
        let converted_case =
            format!("{}", name.ident.as_ref().unwrap().clone()).to_case(Case::UpperCamel);
        let converted_case = format_ident!("{}", converted_case);
        variants.push(converted_case.clone());

        let field_name = name.ident.as_ref().unwrap().clone();
        let field_name_str = format!("{}", field_name);
        field_names.push(quote! { Self::Column::#converted_case => #field_name_str });

        let nn = field_numbers.len() + 1;
        field_numbers.push(quote! { #nn => Self::#converted_case, });

        let ty = &name.ty;
        field_types.push(quote! { <#ty as #microrm_ref::model::Modelable>::column_type() });

        if parse_fk(&name.attrs) {
            let fk_struct_name = format_ident!("{}{}ForeignKey", struct_name, converted_case);
            let ty = &name.ty;
            foreign_keys.push(quote! {
                &#fk_struct_name { col: #enum_name::#converted_case }
            });
            foreign_key_impls.push(quote!{
                struct #fk_struct_name {
                    col: #enum_name
                }
                impl #microrm_ref::entity::EntityForeignKey<#enum_name> for #fk_struct_name {
                    fn local_column(&self) -> &#enum_name { &self.col }
                    fn foreign_table_name(&self) -> &'static str {
                        <<#ty as #microrm_ref::entity::EntityID>::Entity as #microrm_ref::entity::Entity>::table_name()
                    }
                    fn foreign_column_name(&self) -> &'static str {
                        "id"
                    }
                }
            });
        }

        value_references.push(quote! { &self. #field_name });

        index += 1;
        build_clauses.push(quote! { #field_name: <#ty as #microrm_ref::model::Modelable>::build_from(stmt, #index)?.0 });
    }

    let column_types_name = format_ident!(
        "{}_COLUMN_TYPES",
        struct_name.to_string().to_case(Case::ScreamingSnake)
    );

    let field_count = fields.named.iter().count();

    quote!{
        // Related types for #struct_name
        #[derive(Clone,Copy,PartialEq,Hash)]
        #[allow(unused)]
        #[repr(usize)]
        pub enum #enum_name {
            ID,
            #(#variants),*
        }

        #[derive(Debug,PartialEq,Clone,Copy,#microrm_ref::re_export::serde::Serialize,#microrm_ref::re_export::serde::Deserialize)]
        #[allow(unused)]
        pub struct #id_name (i64);

        // Implementations for related types
        impl #microrm_ref::entity::EntityColumns for #enum_name {
            type Entity = #struct_name;
        }

        impl std::convert::From<usize> for #enum_name {
            fn from(i: usize) -> Self {
                match i {
                    0 => Self::ID,
                    #(#field_numbers)*
                    _ => {
                        panic!("Given invalid usize to convert to column")
                    },
                }
            }
        }

        impl #microrm_ref::entity::EntityID for #id_name {
            type Entity = #struct_name;
            fn from_raw_id(raw: i64) -> Self { Self(raw) }
            fn raw_id(&self) -> i64 { self.0 }
        }

        impl #microrm_ref::model::Modelable for #id_name {
            fn bind_to(&self, stmt: &mut #microrm_ref::re_export::sqlite::Statement, col: usize) -> #microrm_ref::re_export::sqlite::Result<()> {
                use #microrm_ref::re_export::sqlite::Bindable;
                self.0.bind(stmt, col)
            }
            fn build_from(stmt: &#microrm_ref::re_export::sqlite::Statement, col_offset: usize) -> #microrm_ref::re_export::sqlite::Result<(Self, usize)> where Self: Sized {
                stmt.read::<i64>(col_offset).map(|x| (#id_name(x), 1))
            }
            fn column_type() -> &'static str where Self: Sized {
                "integer"
            }
        }
        #microrm_ref::re_export::lazy_static::lazy_static!{
            static ref #column_types_name: [&'static str;#field_count + 1] = {
                [
                    "id",
                    #(#field_types),*
                ]
            };
        }

        // Implementations for #struct_name
        impl #microrm_ref::entity::Entity for #struct_name {
            type Column = #enum_name;
            type ID = #id_name;

            fn table_name() -> &'static str { #table_name }
            fn column_count() -> usize {
                // +1 for ID column
                #field_count + 1
            }
            fn index(c: Self::Column) -> usize {
                c as usize
            }
            fn name(c: Self::Column) -> &'static str {
                match c {
                    Self::Column::ID => "ID",
                    #(#field_names),*
                }
            }
            fn values(&self) -> Vec<&dyn #microrm_ref::model::Modelable> {
                vec![ #(#value_references),* ]
            }

            fn build_from(stmt: &#microrm_ref::re_export::sqlite::Statement) -> #microrm_ref::re_export::sqlite::Result<Self> {
                Ok(Self {
                    #(#build_clauses),*
                })
            }

            fn column_types() -> &'static [&'static str] {
                #column_types_name.as_ref()
            }
            fn foreign_keys() -> &'static [&'static dyn #microrm_ref::entity::EntityForeignKey<Self::Column>] {
                &[#(#foreign_keys),*]
            }
        }

        // Foreign key struct implementations
        #(#foreign_key_impls)*
    }.into()
}
