use crate::model::Modelable;

/// A database entity, aka a struct representing a row in a table
pub trait Entity: 'static + for<'de> serde::Deserialize<'de> + serde::Serialize {
    type Column: EntityColumns + 'static + Copy;
    type ID: EntityID;
    fn table_name() -> &'static str;
    fn column_count() -> usize
    where
        Self: Sized;
    fn column_types() -> &'static [&'static str]
    where
        Self: Sized;
    fn index(c: Self::Column) -> usize
    where
        Self: Sized;
    fn name(c: Self::Column) -> &'static str
    where
        Self: Sized;
    fn values(&self) -> Vec<&dyn Modelable>;

    fn build_from(stmt: &sqlite::Statement) -> sqlite::Result<Self>
    where
        Self: Sized;

    fn foreign_keys() -> &'static [&'static dyn EntityForeignKey<Self::Column>];
}

/// Trait representing the columns of a database entity
pub trait EntityColumns: 'static + PartialEq + From<usize> + std::hash::Hash + Clone {
    type Entity: Entity;
}

/// Trait for entity IDs in the database
pub trait EntityID: 'static + std::fmt::Debug + Copy + Modelable {
    type Entity: Entity;
    fn from_raw_id(raw: i64) -> Self;
    fn raw_id(&self) -> i64;
}

/// Trait for a foreign key relationship
pub trait EntityForeignKey<T: EntityColumns> {
    fn local_column(&self) -> &T;
    fn foreign_table_name(&self) -> &'static str;
    fn foreign_column_name(&self) -> &'static str;
}

/// Trait for an index over a column
pub trait Index {
    type IndexedEntity: Entity;

    fn index_name() -> &'static str
    where
        Self: Sized;
    fn columns() -> &'static [<<Self as Index>::IndexedEntity as Entity>::Column]
    where
        Self: Sized;
    fn unique() -> bool
    where
        Self: Sized;
}
